#include <stdio.h>
#include <time.h>
#include <opencv2/opencv.hpp>
#include <ippi.h>
#include <ippcc.h>

using namespace cv;

char helper[] = 
  "01_OpenCVvsIPP.exe <img_name> <mode> <num_of_exp>\n\
  \t<img_name> - image filename\n\
  \t<mode>:\n\
  \t\t1 - median filtering\n\
  \t\t2 - erode\n\
  \t\t3 - dilate\n\
  \t\t4 - calc histogram\n\
  \t<num_of_exp> - number of experiments\n";

double median_opencv(const Mat &srcImg, Mat &dstImg, const int kSize = 3);
double erode_opencv(const Mat &srcImg, Mat &dstImg);
double dilate_opencv(const Mat &srcImg, Mat &dstImg);
double hist_opencv(const Mat &srcImg, Mat &dstImg);

double median_ipp(const Mat &srcImg, Mat &dstImg, const int msk = 3);
double erode_ipp(const Mat &srcImg, Mat &dstImg);
double dilate_ipp(const Mat &srcImg, Mat &dstImg);
double hist_ipp(const Mat &srcImg, Mat &dstImg);

int compare(const Mat &srcImgOCV, const Mat &srcImgIPP);

double median_opencv(const Mat &srcImg, Mat &dstImg, const int kSize)
{
  clock_t start, finish;

  start = clock();
  medianBlur(srcImg, dstImg, kSize);
  finish = clock();

  return double(finish - start) / CLOCKS_PER_SEC;
}

double erode_opencv(const Mat &srcImg, Mat &dstImg)
{
  clock_t start, finish;

  start = clock();
  Mat element = Mat();
  erode(srcImg, dstImg, element);
  finish = clock();

  return double(finish - start) / CLOCKS_PER_SEC;
}

double dilate_opencv(const Mat &srcImg, Mat &dstImg)
{
  clock_t start, finish;

  start = clock();
  Mat element = Mat();
  dilate(srcImg, dstImg, element);
  finish = clock();

  return double(finish - start) / CLOCKS_PER_SEC;
}

double hist_opencv(const Mat &srcImg, Mat &dstImg)
{
  Mat bgrChannels[3], bHist, gHist, rHist;
  //   
  int kBins = 256;
  //    
  float range[] = {0.0f, 256.0f};
  const float* histRange = { range };
  //     
  bool uniform = true; 
  //     
  bool accumulate = false; 
  clock_t start, finish;

  start = clock();
  split(srcImg, bgrChannels);
  //     
  calcHist(&bgrChannels[0], 1, 0, Mat(), bHist, 1, 
    &kBins, &histRange, uniform, accumulate);
  calcHist(&bgrChannels[1], 1, 0, Mat(), gHist, 1, 
    &kBins, &histRange, uniform, accumulate);
  calcHist(&bgrChannels[2], 1, 0, Mat(), rHist, 1, 
    &kBins, &histRange, uniform, accumulate);
  finish = clock();

  return double(finish - start) / CLOCKS_PER_SEC;
}

double median_ipp(const Mat &srcImg, Mat &dstImg, const int msk)
{
  const int borderSize = msk / 2;
  IppiSize mskSize = { msk , msk };
  IppiPoint anchor = { msk / 2, msk / 2 };
  Mat ippSrcImg, ippDstImg;
  Size ippImgSize;
  IppiSize ippSrcSize, ippDstSize;
  Ipp8u *pSrcData, *pDstData;
  clock_t start, finish;

  ippImgSize.width  = srcImg.size().width  + 2 * borderSize;
  ippImgSize.height = srcImg.size().height + 2 * borderSize;
  ippSrcImg.create(ippImgSize, srcImg.type());
  ippDstImg.create(ippImgSize, srcImg.type());
  pSrcData = (Ipp8u *)ippSrcImg.data;
  pDstData = (Ipp8u *)ippDstImg.data;

  start = clock();
  ippSrcSize.width  = srcImg.size().width;
  ippSrcSize.height = srcImg.size().height;
  ippDstSize.width  = srcImg.size().width  + 2 * borderSize;
  ippDstSize.height = srcImg.size().height + 2 * borderSize;
  ippiCopyReplicateBorder_8u_C3R(srcImg.data, srcImg.step1(), ippSrcSize, 
    pSrcData, srcImg.step1() + 3 * 2 * borderSize, ippDstSize, borderSize, 
    borderSize);

  ippiFilterMedian_8u_C3R(
    pSrcData + ippImgSize.width * 3 * borderSize + 3 * borderSize, 
    srcImg.step1() + 3 * 2 * borderSize, 
    pDstData + ippImgSize.width * 3 * borderSize + 3 * borderSize, 
    srcImg.step1() + 3 * 2 * borderSize, ippSrcSize, mskSize, anchor);
  finish = clock();

  dstImg.create(srcImg.size(), srcImg.type());
  pSrcData = pDstData;
  pDstData = (Ipp8u *)dstImg.data;
  ippiCopy_8u_C3R(pSrcData + ippImgSize.width * 3 * borderSize + 3 * borderSize, 
    srcImg.step1() + 3 * 2 * borderSize, pDstData, dstImg.step1(), ippSrcSize);

  ippSrcImg.release();
  ippDstImg.release();

  return double(finish - start) / CLOCKS_PER_SEC;
}

double erode_ipp(const Mat &srcImg, Mat &dstImg)
{
  IppiMaskSize msk;
  Ipp8u *pSrcData, *pDstData;
  IppiSize srcSize, dstRoiSize;
  clock_t start, finish;

  srcImg.copyTo(dstImg);

  start = clock();
  pSrcData = (Ipp8u *)srcImg.data;
  pDstData = (Ipp8u *)dstImg.data;
  srcSize.width = srcImg.size().width;
  srcSize.height = srcImg.size().height;

  msk = ippMskSize3x3;
  dstRoiSize.width = srcImg.size().width - 2;
  dstRoiSize.height = srcImg.size().height - 2;
  ippiErode3x3_8u_C3R(pSrcData + srcSize.width * 3 * 1 + 3 * 1, srcImg.step1(), 
    pDstData + srcSize.width * 3 * 1 + 3 * 1, srcImg.step1(), dstRoiSize);

  finish = clock();

  return double(finish - start) / CLOCKS_PER_SEC;
}

double dilate_ipp(const Mat &srcImg, Mat &dstImg)
{
  IppiMaskSize msk;
  Ipp8u *pSrcData, *pDstData;
  IppiSize srcSize, dstRoiSize;
  clock_t start, finish;

  srcImg.copyTo(dstImg);

  start = clock();
  pSrcData = (Ipp8u *)srcImg.data;
  pDstData = (Ipp8u *)dstImg.data;
  srcSize.width = srcImg.size().width;
  srcSize.height = srcImg.size().height;

  msk = ippMskSize3x3;
  dstRoiSize.width = srcImg.size().width - 2;
  dstRoiSize.height = srcImg.size().height - 2;
  ippiDilate3x3_8u_C3R(pSrcData + srcSize.width * 3 * 1 + 3 * 1, srcImg.step1(), 
    pDstData + srcSize.width * 3 * 1 + 3 * 1, srcImg.step1(), dstRoiSize);

  finish = clock();

  return double(finish - start) / CLOCKS_PER_SEC;
}

double hist_ipp(const Mat &srcImg, Mat &dstImg)
{
  //   
  const int kBins = 256;
  Ipp8u *pSrcData;
  IppiSize srcSize;
  int i;
  Ipp32s level1[kBins + 1], level2[kBins + 1], level3[kBins + 1];
  Ipp32s *pHisto[3];
  const Ipp32s *pLevels[3] = { level1, level2, level3 };
  int nLevels[3] = {kBins + 1, kBins + 1, kBins + 1};
  clock_t start, finish;

  pHisto[0] = new Ipp32s[kBins + 1];
  pHisto[1] = new Ipp32s[kBins + 1];
  pHisto[2] = new Ipp32s[kBins + 1];

  for (i = 0; i < kBins + 1; i++)
    level1[i] = level2[i] = level3[i] = i;

  pSrcData = (Ipp8u *)srcImg.data;
  srcSize.width = srcImg.size().width;
  srcSize.height = srcImg.size().height;

  start = clock();
  ippiHistogramRange_8u_C3R(pSrcData, srcImg.step1(), srcSize, pHisto, pLevels, nLevels);
  finish = clock();

  delete [] pHisto[0];
  delete [] pHisto[1];
  delete [] pHisto[2];

  return double(finish - start) / CLOCKS_PER_SEC;
}

int compare(const Mat &srcImgOCV, const Mat &srcImgIPP)
{
  int kPoints = 0;
  int w, h, i, j, idx;
  w = srcImgOCV.size().width;
  h = srcImgOCV.size().height;
  for (i = 0; i < h; i++)
  {
    idx = 3 * i * w;
    for (j = 0; j < w; j++, idx += 3)
    {
      if (((srcImgOCV.data[idx] - srcImgIPP.data[idx]) != 0) 
        ||((srcImgOCV.data[idx + 1] - srcImgIPP.data[idx + 1]) != 0)
        ||((srcImgOCV.data[idx + 2] - srcImgIPP.data[idx + 2]) != 0))
      {
        kPoints++;
      }
    }
  }

  return kPoints;
}

int main(int argc, char *argv[])
{
  Mat srcImgOCV, srcImgIPP, dstImgOCV, dstImgIPP;
  int mode, i;
  int kDiffPoints = 0;
  int numOfExp;
  double ocv_time, ipp_time;

  if (argc < 4)
  {
    printf("%s", helper);
    return 1;
  }

  //  
  srcImgOCV = imread(argv[1]);
  if (srcImgOCV.data == 0)
  {
    printf("ERROR!!! imread(...)");
    return 1;
  }
  srcImgOCV.copyTo(srcImgIPP);

  mode = atoi(argv[2]);
  numOfExp = atoi(argv[3]);

  ocv_time = ipp_time = 0;
  switch (mode)
  {
  case 1:
    for (i = 0; i < numOfExp; i++)
    {
      //    OpenCV
      ocv_time += median_opencv(srcImgOCV, dstImgOCV);
      //    IPP
      ipp_time += median_ipp(srcImgIPP, dstImgIPP);
    }
    break;
  case 2:
    for (i = 0; i < numOfExp; i++)
    {
      //     OpenCV
      ocv_time += erode_opencv(srcImgOCV, dstImgOCV);
      //     IPP
      ipp_time += erode_ipp(srcImgIPP, dstImgIPP);
    }
    break;
  case 3:
    for (i = 0; i < numOfExp; i++)
    {
      //     OpenCV
      ocv_time += dilate_opencv(srcImgOCV, dstImgOCV);
      //     IPP
      ipp_time += dilate_ipp(srcImgIPP, dstImgIPP);
    }
    break;
  case 4:
    for (i = 0; i < numOfExp; i++)
    {
      //     OpenCV
      ocv_time += hist_opencv(srcImgOCV, dstImgOCV);
      //     IPP
      ipp_time += hist_ipp(srcImgIPP, dstImgIPP);
    }
    break;
  }

  //   
  if (mode < 4)
  {
    kDiffPoints = compare(dstImgOCV, dstImgIPP);
    if (kDiffPoints == 0)
      printf("OpenCV and IPP give the same result.\n");
    else
      printf("OpenCV and IPP give different results in %d points.\n", kDiffPoints);
  }
  printf("ocv time is %.3f\nipp time is %.3f", ocv_time, ipp_time);
  /*
  //  
  namedWindow("srcImg");
  imshow("srcImg", srcImgOCV);
  namedWindow("dstImgOCV");
  imshow("dstImgOCV", dstImgOCV);
  namedWindow("dstImgIPP");
  imshow("dstImgIPP", dstImgIPP);
  waitKey();

  //  
  destroyAllWindows();
  */
  //  
  srcImgOCV.release();
  srcImgIPP.release();
  dstImgOCV.release();
  dstImgIPP.release();

  return 0;
}

